/*
 * userspace_listner.c - Listening in userspace
 * 
 *Acknowledgement -> ELSA (http://elsa.sourceforge.net)
 */

#include <assert.h>		/* for assertion */
#include <dirent.h>		/* opendir() */
#include <errno.h>		/* errno */
#include <fcntl.h>		/* open(), opendir() */
#include <signal.h>		/* sigaction() */
#include <stdio.h>		/* perror() */
#include <stdlib.h>		/* exit() */
#include <string.h>		/* strerror() */
#include <errno.h>		/* sys_errlist */
#include <syslog.h>		/* syslog() */
#include <unistd.h>		/* sysconf(), setsid(), fork(), dup(), close() */
#include <selinux/selinux.h>
#include <asm/types.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>		/* fork(), open(), waitpid(), IPC */
#include <sys/un.h>

#include <linux/netlink.h>
/* 
 * As connector is not included in the libc we need to include it through
 * kernel source. The problem is that there is a conflit with some 
 * definitions in time.h. Some variable are defined in sys/time.h and
 * redefined in linux/time.h (included in cn_proc.h). Thus we add this
 * ugly definition to be able to compile. We will remove that hack as 
 * soon as libc includes connectors.
 */
#define _LINUX_TIME_H
#include <linux/cn_proc.h>
#include <linux/connector.h>
#undef _LINUX_TIME_H

#include "userspace_classifier.h"




static int sk_nl ;
static int process_pid ;
static void cleanup()
{
	close(sk_nl);
	exit(EXIT_SUCCESS);
}

/* Handles events on process like FORK and EXEC 
In case of FORK : normally selinux security context don't change in fork system call, 
			So, we can ignore this system call
In case of EXEC : Find out the security context of current process, and
		determine, to which class it should belong... 
		This will be done by use of function search_security_context () function.
*/
int handle_event (int pid, int type )
{

	int ret ;
	security_context_t s_context ;
	struct security_context_list *search_result;
	char * cpuset_name ;

	
	s_context = NULL ;
	ret = getpidcon ( pid, &s_context ) ;
	if ( ret == -1 )
	{
		syslog(LOG_INFO, "Exec: pid=[%d], Error: getpidcon failed : %s ", process_pid, strerror(errno) );
		return -1 ;
	}

	switch ( type )
	{

	case PROC_EVENT_FORK :
		syslog(LOG_INFO, "Request for fork : pid %d | %s |\n",  process_pid, s_context );
		break ;
		
	case PROC_EVENT_EXEC :

		// find the security context of current process and classify it
		search_result = search_security_context ( s_context ) ;
		if (search_result == NULL )
		{
			syslog(LOG_INFO, 
			  "Exec: pid=[%d], context=[%s], class=[default classification], action=[no action taken]\n",
			  process_pid, s_context );
			return 0;
		}
		
		// now, create the cpuset with given name.
		cpuset_name = search_result->cpuset->data->cpuset_name ;
		
		// Add current process to searched cpuset class.
		ret = add_member_to_class(process_pid,search_result);			
		if(ret == -1)
		{
			syslog(LOG_INFO, 
			  "Exec: pid=[%d], context=[%s], class=[%s], action=[Error in adding to class]\n",
			  process_pid, s_context, cpuset_name );
			return -1 ;
		}
		syslog(LOG_INFO, 
		  "Exec: pid=[%d], context=[%s], class=[%s], action=[Added]\n",
		  process_pid, s_context, cpuset_name );
		break ;
	
	} // end switch : type
	
	// Free the memory alloted by kernel to s_context
	// which ia alloted internally by getpidcon () function
	freecon (s_context) ;
	return 0 ;
} // end function : 


/*
This is an infinite loop, receiving messages from kernel on netlink socket

*/
static void recv_in_loop( int sk )
{
	char buff[1024];	/* it's large enough */
	struct nlmsghdr *hdr;
	struct cn_msg *msg;
	struct proc_event *ev;
	int len;
	int cpuid;
	int ppid;
	int ptgid;
	int cpid;
	int ctgid;



	for (;;)
	{
		/* Clear the buffer */
		memset(buff, '\0', sizeof(buff));

		/* Listen */
		len = recv(sk, buff, sizeof(buff), 0);
		if (len == -1) {
			syslog(LOG_INFO, "netlink recv error");
			continue ;
		}

		/* point to the message header */
		hdr = (struct nlmsghdr *)buff;

		switch (hdr->nlmsg_type) {
		case NLMSG_DONE:
			msg = (struct cn_msg *)NLMSG_DATA(hdr);
			ev = (struct proc_event *)msg->data;
			switch (ev->what )
			{
		
			case PROC_EVENT_FORK :
				cpuid = ev->cpu;
				process_pid  = ev->event_data.fork.child_pid;

				handle_event (process_pid, ev->what ) ;	

				break ;
			
			case PROC_EVENT_EXEC :
				cpuid = ev->cpu;
				process_pid  = ev->event_data.exec.process_pid;

				handle_event (process_pid, ev->what ) ;	
				break ;

			default :
//				printf ("Other netlink message\n");
				break ;

			} // end switch : ev->what

			break;


		case NLMSG_ERROR:
			syslog(LOG_INFO, "recv_sk_nl error");
			/* Fall through */
		default:
			break;

		} // end switch  (hdr->nlmsg_type)
	} // end for :

	return;
} // end function :

/**
* closeall - close all FDs >= a specified value 
* @fd: a file descriptor
*
* close all FDs wich are greater or equal to a system value
*/
static void closeall(int fd) 
{
	int fdlimit = sysconf(_SC_OPEN_MAX);
	while (fd < fdlimit)
		close(fd++);
}

int daemonize(int nochdir, int noclose)
{
	/* 
	 * First, we fork to allow parent to exit. This returns control to 
	 * the command line. 
	 */
	switch (fork()) {
	case -1:
		/* error */
		return -1;
	case 0:
		/* child */
		break;
	default:
		/* Parent can exit */
		exit(EXIT_SUCCESS);
	}

	/* Now we're not a process group leader */
	if (setsid() < 0)	/* shoudn't fail */
		return -1;

	/* 
	 * New session is created and we are the leader. 
	 * Since a  controlling terminal is associated with a session, 
	 * and this new session has not yet acquired a controlling terminal 
	 * our process now has no controlling terminal, which is a Good Thing 
	 * for daemons.
	 * fork() again so the parent, (the session group leader), can exit.
	 * This means that we, as a non-session group leader, can never regain a
	 * controlling terminal.
	 */
	switch (fork()) {
	case -1:
		/* error */
		return -1;
	case 0:
		/* child */
		break;
	default:
		/* Parent can exit */
		exit(EXIT_SUCCESS);
	}

	umask(0);


	if (!noclose) {
		closeall(0);
		open("/dev/null", O_RDWR);
		open (LOG_FILE_PATH, O_WRONLY|O_CREAT|O_APPEND,0640 );
		dup(1);
		dup(1);
	}

	if (!nochdir)
		chdir("/");
	return 0;
}


/**
 * main - Principal routine and also a river rising in eastern Germany.
 *
 * Here are the different tasks:
 *    1) Open a log file
 *    2) Read configuration rules to be applied
 *    3) Become a daemon
 *    4) Install signal handler
 *    5) Bind to netlink socket
 *    6) Listen on it in infinite loop ( recv_in_loop() )
 */

int main(int argc, char **argv)
{
	int err;
	struct sockaddr_nl sa_nl;	/* info for the netlink interface */


	// Open connection to system logger
	//openlog (log_ident,LOG_PID,LOG_INFO);

	// read the rules given by user for automatic classification
	 file_parse(CONTAINER_RULEBASE_PATH);
	 security_context_file_parser(CLASSIFICATION_RULEBASE_PATH);




	if (daemonize(0, 0) < 0) {
		perror("daemonize");
		exit(EXIT_FAILURE);
	}





	/* Install signal handler */
	signal(SIGUSR1, cleanup);

	/*
	 * The netlink communication 
	 * 
	 * Create an endpoint for communication. Use the kernel user
	 * interface device (PF_NETLINK) which is a datagram oriented
	 * service (SOCK_DGRAM). The netlink family used is
	 * NETLINK_CONNECTOR because we use the Process Events Connector
	 * patch.
	 */
	sk_nl = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_CONNECTOR);
	if (sk_nl == -1) 
	{
		syslog(LOG_INFO, "socket sk_nl error");
		return -1;
	}

	sa_nl.nl_family = AF_NETLINK;
	sa_nl.nl_groups = CN_IDX_PROC;
	sa_nl.nl_pid    = getpid();

	err = bind(sk_nl, (struct sockaddr *)&sa_nl, sizeof(sa_nl));
	if (err == -1) 
	{
		syslog(LOG_INFO, "binding sk_nl error");
		close(sk_nl);
		return -1;
	}

	

	// infinite loop, accepting netlink messages from kernel
	recv_in_loop (sk_nl);
	
	return 0 ;

}
